#include "common.h"
#include "uart.h"
#include "rtsp.h"

// Configuration registers for FX/FS MIDI adapter

_FICD(ICS_PGD2 & JTAGEN_OFF)
_FPOR(ALTI2C1_OFF & ALTI2C2_OFF & WDTWIN_WIN25)
_FWDT(WDTPOST_PS32768 & WDTPRE_PR128 & PLLKEN_OFF & WINDIS_OFF & FWDTEN_OFF)
_FOSC(POSCMD_NONE & OSCIOFNC_ON & IOL1WAY_OFF & FCKSM_CSECMD)
_FOSCSEL(FNOSC_FRC & IESO_OFF)
_FGS(GWRP_OFF & GCP_OFF)

#define MAGIC_BYTES "\x02\x0F\x0F\x14"	//"BOOT" in control chars
//define MAGIC_BYTES "boot"
#define MAGIC_RESPONSE 0x06180613		//"FXFS" in control chars, big-endian

const uint16 __attribute__((space(psv),__section__((".config")))) config_word = 0xbad;

//const uint16 __attribute__((space(psv),__section__((".test")))) test_word = 0x1234;

void init() {
	ANSELA = CONFIG_ANSELA;
	ANSELB = CONFIG_ANSELB;
	LATA = UART_TX_ON;
	LATB = 0;
	TRISA = CONFIG_TRISA;
	TRISB = CONFIG_TRISB;
	uart_init();
}

void wait_100ms() {
	//Initialize timer1 (16-bit) for 0.1 sec operation at 7.37MHz
	TMR1 = 0;
	PR1 = 1440;
	_T1IF = 0;
	T1CON = 0x8030;
	while(!_T1IF) { }
}

uint32 read_response() {
	uint32 value = 0;
	while(_URXDA) {
		value <<= 8;
		value |= (U1RXREG & 0xFF);
	}
	return value;
}

void get_firmware();
bool is_firmware_good() { return (rtsp_ReadConfigWord() == 0x600d); }

void test_rtsp();

void blink_error() {
	uint16 blinks = 0;
	uint16 space = 0;
	TMR1 = 0;
	_T1IF = 0;
	while(true) {
		while(!_T1IF) { }
		_T1IF = 0;
		
		if(blinks) {
			LATB ^= LED_ON;
			--blinks;
		} else if(space) {
			--space;
		} else {
			blinks = 6;
			space = 10;
		}
		if(_URXDA) {
			/*if(U1RXREG == 'r') {
				uart_send("wr",2);
				test_rtsp();
			}
			else */ asm volatile("reset");
		}
	}
}

void send_ack() { uart_send("*",1); }
void send_nak() { uart_send("!",1); }

int main(void) {
	init();

	wait_100ms();
	uart_send(MAGIC_BYTES,4);
	wait_100ms();
	if(read_response() == MAGIC_RESPONSE) {
		send_ack();
		get_firmware();
	}

	if(is_firmware_good()) {
		asm volatile("goto 0x200");
		//uart_send("good",4);
	} else {
		blink_error();
	}
	while(true) Idle();
}

byte page_buffer[3072];

byte read_byte() {
	uint16 timeout = 5;
	TMR1 = 0;
	while(timeout) {
		_T1IF = 0;
		while(!_T1IF) {
			if(_URXDA) return (byte)U1RXREG;
		}
		if(_T1IF) --timeout;
	}
	asm volatile("reset");
	return 0;
}

uint16 read_page_size() {
	uint16 page_size;
	do {
		page_size = read_byte();
		if(page_size <= 16)
			return page_size;
		send_nak();
	} while(true);
}

void blink_waiting() {
	TMR1 = 0;
	while(!_URXDA) {
		LATB ^= LED_ON;
		_T1IF = 0;
		while(!_T1IF && !_URXDA) { }
	}
}

void set_firmware_bad() {
	if(is_firmware_good()) {
		rtsp_WriteConfigWord(0x0bad);
	}
}

void set_firmware_good() {
	if(!is_firmware_good()) {
		rtsp_WriteConfigWord(0x600d);
	}
}

bool check_checksum(byte checksum) {
	if((checksum + read_byte()) & 0xFF) {
		return false;
	} else {
		return true;
	}
}

byte read_page(uint16 n_bytes) {
	byte checksum = 0;
	byte temp = 0;
	byte* pageptr = page_buffer;
	uint16 i;
	for(i = 0; i < n_bytes; ++i,++pageptr) {
		temp = read_byte();
		checksum += temp;
		*pageptr = temp;
	}
	return checksum;
}

void get_firmware() {
	//Serial protocol:
	//struct {
	//	byte page_size_in_lines (0..16 - each line is 192 bytes / 64 instructions; not distinguished by hardware)
	//  byte page_data[192*page_size_in_lines]	(24-bit little-endian words)
	//	byte checksum (= negative of simple sum of all page_data bytes)
	//} pages[21]

	//Ack/nak after page size and checksum

	//500ms timeout anywhere except first byte => resets PIC

	//Blink LED until a byte is received
	//For each page:
	//	Read page size and nak until page size is good
	//  Set firmware bad
	//  ack
	//	If page size is nonzero, read all data bytes in page
	//	Read and check checksum; if bad, nak and go back to beginning of this page.
	//	Write the page
	//  ack
	//Set firmware good

	uint16 page;
	uint16 page_size;
	byte checksum;

	blink_waiting();
	for(page = 0; page <= RTSP_PAGE_VAR; ++page) {
		while(true) {
			page_size = read_page_size() * 192;
			set_firmware_bad();
			send_ack();
			if(page_size) {
				checksum = read_page(page_size);
			} else {
				checksum = 0;
			}
			if(check_checksum(checksum)) {
				rtsp_WritePage(page,page_buffer,page_size);
				send_ack();
				break; //while
			} else {
				send_nak();
			}
		}
	}
	set_firmware_good();
}

void test_rtsp() {
	rtsp_WriteConfigWord(0xbad0);
	uart_send("~~",2);
}
